#!/usr/bin/env python2
# -*- coding: utf-8 -*-
from __future__ import print_function
from .__init__ import *


"""r0c.py: retr0chat Telnet/Netcat Server"""
__author__    = "ed <a@ocv.me>"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2018


import os
import sys
import signal

if not 'r0c' in sys.modules:
	print('\r\n  retr0chat must be launched as a module.\r\n  in the project root, run this:\r\n\r\n    python -m r0c\r\n')
	sys.exit(1)

from .config   import *
from .util     import *
from .ivt100   import *
from .inetcat  import *
from .itelnet  import *
from .chat     import *
from .user     import *
from .world    import *

if not PY2:
	from .diag import *


class Core(object):
	def __init__(self):
		pass

	def start(self, args=None):
		if args is None:
			args = sys.argv
		
		if len(args) < 3:
			print()
			print('  need argument 1:  Telnet port  (or 0 to disable)')
			print('  need argument 2:  NetCat port  (or 0 to disable)')
			print('  optional arg. 3:  Password')
			print()
			print('  example 1:')
			print('    python -m r0c 2323 1531 hunter2')
			print()
			print('  example 2:')
			print('    python -m r0c 23 531')
			print()
			return False

		for d in ['pm','chan','wire']:
			try: os.makedirs(EP.log + d)
			except: pass

		print('  *  r0c {0}, py {1}'.format(S_VERSION, host_os()))

		self.telnet_port = int(args[1])
		self.netcat_port = int(args[2])

		print('  *  Telnet server on port ' + str(self.telnet_port))
		print('  *  NetCat server on port ' + str(self.netcat_port))

		if not self.read_password(args):
			return False
		
		print('  *  Logs at ' + EP.log)
		compat_chans_in_root()

		self.stopping = 0
		self.threadmon = False
		self.pushthr_alive = False
		self.asyncore_alive = False

		print('  *  Capturing ^C')
		signal.signal(signal.SIGINT, self.signal_handler)

		print('  *  Creating world')
		self.world = World(self)

		print('  *  Starting Telnet server')
		self.telnet_server = TelnetServer('0.0.0.0', self.telnet_port, self.world, self.netcat_port)

		print('  *  Starting NetCat server')
		self.netcat_server = NetcatServer('0.0.0.0', self.netcat_port, self.world, self.telnet_port)

		print('  *  Loading user configs')
		self.telnet_server.load_configs()
		self.netcat_server.load_configs()

		print('  *  Starting push driver')
		self.push_thr = threading.Thread(target=self.push_worker, args=(
			self.world, [self.telnet_server, self.netcat_server],), name='push')
		#self.push_thr.daemon = True
		self.push_thr.start()

		print('  *  Handover to asyncore')
		self.asyncore_thr = threading.Thread(target=self.asyncore_worker, name='ac_mgr')
		self.asyncore_thr.start()

		return True


	def read_password(self, args):
		self.password = ADMIN_PWD
		
		# password as argument overrides all others
		if len(args) > 3:
			self.password = args[3]
			print('  *  Password from argument')
			return True
		
		# password file in home directory overrides config
		pwd_file = os.path.join(EP.app, 'password.txt')
		if os.path.isfile(pwd_file):
			print('  *  Password from ' + pwd_file)
			with open(pwd_file, 'rb') as f:
				self.password = f.read().decode('utf-8').strip()
				return True

		# fallback to config.py
		print('  *  Password from ' + os.path.join(EP.src, 'config.py'))
		if self.password != u'hunter2':
			return True
		
		# default password is verboten
		print()
		print('\033[1;31m  change the ADMIN_PWD in the path above \033[0m')
		print('\033[1;31m  or provide your password as an argument \033[0m')
		print('\033[1;31m  or save it here: ' + pwd_file + '\033[0m')
		print()
		return False


	def run(self):
		print('  *  r0c is up')
		
		if not BENCHMARK:
			while not self.stopping:
				time.sleep(0.1)
		else:
			last_joins = 0
			last_parts = 0
			last_messages = 0
			while not self.stopping:
				for n in range(20):
					if self.stopping:
						break
					time.sleep(0.1)

				print('{0:.3f}  j {1}  p {2}  m {3}  d {4},{5},{6}'.format(time.time(),
					self.world.num_joins, self.world.num_parts, self.world.num_messages,
					self.world.num_joins - last_joins,
					self.world.num_parts - last_parts,
					self.world.num_messages - last_messages))
				
				last_joins = self.world.num_joins
				last_parts = self.world.num_parts
				last_messages = self.world.num_messages

		print('\r\n  *  asyncore stopping')
		clean_shutdown = False
		for n in range(0, 40):  # 2sec
			if not self.asyncore_alive:
				print('  *  asyncore stopped')
				clean_shutdown = True
				break
			time.sleep(0.05)
		
		if not clean_shutdown:
			print(' -X- asyncore is stuck')

		print('  *  Saving user configs')
		self.telnet_server.save_configs()
		self.netcat_server.save_configs()

		print('  *  asyncore cleanup')
		self.netcat_server.close()
		self.telnet_server.close()
		
		print('  *  r0c is down')


	def asyncore_worker(self):
		self.asyncore_alive = True

		timeout = 0.05
		while not self.stopping:
			try:
				asyncore.loop(timeout, count=0.5/timeout)
			except Exception as ex:
				if 'Bad file descriptor' in str(ex):
					#print('osx bug ignored')
					continue
				whoops()

		self.asyncore_alive = False


	def push_worker(self, world, ifaces):
		self.pushthr_alive = True
		
		nth_iter = 0
		last_ts = None
		last_date = None
		while not self.stopping:
			while True:
				ts = time.time()
				its = int(ts)
				if its != last_ts:
					last_ts = its
					#print('=== {0}'.format(its))
					break
				if ts - its < 0.98:
					#print(ts-its)
					time.sleep((1-(ts-its))*0.9)
				else:
					time.sleep(0.01)

			with world.mutex:
				#ts = (ts - 1516554584) * 10000
				date = datetime.datetime.utcfromtimestamp(ts).strftime('%Y-%m-%d')
				if date != last_date:
					if last_date:
						world.broadcast_message(
							u"\033[36mday changed to \033[1m{0}".format(date), False)
					last_date = date

				for iface in ifaces:
					for client in iface.clients:
						if client.handshake_sz:
							client.refresh(False)

					#print('next scheduled kick: {0}'.format('x' if iface.next_scheduled_kick is None else iface.next_scheduled_kick - ts))

					if iface.next_scheduled_kick is not None \
					and iface.next_scheduled_kick <= ts:
						to_kick = []
						next_min = None
						for sch in iface.scheduled_kicks:
							if sch[0] <= ts:
								to_kick.append(sch)
							else:
								if next_min is None \
								or next_min > sch[0]:
									next_min = sch[0]
						
						for sch in to_kick:
							timeout, remote, msg = sch
							iface.scheduled_kicks.remove(sch)
							if remote in iface.clients:
								if msg is None:
									iface.part(remote)
								else:
									iface.part(remote, False)
									print(msg)

						iface.next_scheduled_kick = next_min

				nth_iter += 1
				if nth_iter % 600 == 0:

					# flush client configs
					for iface in ifaces:
						iface.save_configs()

						# flush wire logs
						if LOG_RX or LOG_TX:
							for client in iface.clients:
								if client.wire_log:
									try: client.wire_log.flush()
									except: whoops()

					# flush chan logs
					for chan_list in [world.pub_ch, world.priv_ch]:
						for chan in chan_list:
							if chan.log_fh:
								try: chan.log_fh.flush()
								except: whoops()

		self.pushthr_alive = False


	def shutdown(self):
		#monitor_threads()
		self.stopping += 1
		if self.stopping >= 3:
			os._exit(1)


	def signal_handler(self, signal, frame):
		if THREADMON and not self.threadmon:
			self.threadmon = True
			monitor_threads()
		else:
			self.shutdown()


def start_r0c(args):
	core = Core()
	try:
		if core.start(args):
			core.run()
	except:
		whoops()
		os._exit(1)


def main(args=None):
	mode = 'normal'
	#mode = 'profiler'
	#mode = 'unrag-speedtest'
	#mode = 'unrag-layout-test-v1'
	#mode = 'unrag-layout-test-interactive'
	#mode = 'test-ansi-annotation'
	#test_hexdump()


	if mode == 'normal':
		start_r0c(args)

	if mode == 'profiler':
		print('  *  PROFILER ENABLED')
		statfile = 'profiler-results'
		import yappi
		yappi.start()
		start_r0c(args)
		yappi.stop()
		
		fn_stats = yappi.get_func_stats()
		thr_stats = yappi.get_thread_stats()

		print()
		for ext in ['pstat','callgrind','ystat']:
			print('writing {0}.{1}'.format(statfile, ext))
			fn_stats.save('{0}.{1}'.format(statfile, ext), type=ext)

		with open('{0}.func'.format(statfile), 'w') as f:
			fn_stats.print_all(out=f)

		with open('{0}.thr'.format(statfile), 'w') as f:
			thr_stats.print_all(out=f)

		print('\n\n{0}\n  func stats\n{0}\n'.format('-'*72))
		fn_stats.print_all()
		
		print('\n\n{0}\n  thread stats\n{0}\n'.format('-'*72))
		thr_stats.print_all()


	if mode == 'unrag-speedtest':
		bench_unrag('../radio.long')

	if mode == 'unrag-layout-test-v1':
		unrag_layout_test_dump()

	if mode == 'unrag-layout-test-interactive':
		unrag_layout_test_interactive()

	if mode == 'test-ansi-annotation':
		test_ansi_annotation()


if __name__ == '__main__':
	main()



# -*- coding: utf-8 -*-
from __future__ import print_function
from .__init__ import *
if __name__ == '__main__':
	raise RuntimeError('\r\n{0}\r\n\r\n  this file is part of retr0chat.\r\n  enter the parent folder of this file and run:\r\n\r\n    python -m r0c <telnetPort> <netcatPort>\r\n\r\n{0}'.format('*'*72))

import re
import os
import datetime

from .util import *
from .chat import *

if PY2:
	from Queue import Queue
else:
	from queue import Queue



class World(object):
	def __init__(self, core):
		self.core = core
		self.users = []            # User instances
		self.pub_ch = []           # NChannel instances (public)
		self.priv_ch = []          # NChannel instances (private)
		self.dirty_ch = {}         # Channels that have pending tx
		self.task_queue = Queue()  # Delayed processing of expensive tasks
		self.mutex = threading.RLock()

		# config
		self.messages_per_log_file = MESSAGES_PER_LOG_FILE

		# stats for benchmarking
		self.num_joins = 0
		self.num_parts = 0
		self.num_messages = 0

		threading.Thread(target=self.refresh_chans, name='tx_chan').start()


	def add_user(self, user):
		with self.mutex:
			self.users.append(user)


	def find_user(self, nick):
		with self.mutex:
			nick = nick.lower()
			for usr in self.users:
				if usr.nick and usr.nick.lower() == nick:
					return usr
			return None


	def refresh_chans(self):
		self.chan_sync_active = True
		while not self.core.stopping:
			time.sleep(0.05)
			with self.mutex:
				#while not self.task_queue.empty():
				#	task = self.task_queue.get()
				#	task[0](*task[1],**task[2])

				dirty_ch = list(self.dirty_ch)
				self.dirty_ch = {}

				for chan in dirty_ch:
					self.refresh_chan(chan)

		self.chan_sync_active = False


	def refresh_chan(self, nchan):
		if not nchan.uchans:
			# all users left since this channel got scheduled for refresh
			return

		last_msg = nchan.msgs[-1].sno if nchan.msgs else 0
		
		for uchan in nchan.uchans:
			uchan.update_activity_flags(False, last_msg)
			if uchan.user.active_chan == uchan:
				if not uchan.user.client.handshake_sz \
				or uchan.user.client.wizard_stage is not None:
					continue

				#print('refreshing {0} for {1}'.format(nchan.get_name(), uchan.user.nick))
				uchan.user.client.refresh(False)


	def send_chan_msg(self, from_nick, nchan, text, ping_self=True):
		max_hist_mem = MAX_HIST_MEM
		msg_trunc_size = MSG_TRUNC_SIZE
		with self.mutex:
			self.num_messages += 1
			if nchan.name is None and not from_nick.startswith(u'-'):
				# private chan, check if we have anyone to send to
				if len(nchan.uchans) == 1:
					if nchan.uchans[0].alias == u'r0c-status':
						if nchan.uchans[0].user.nick == from_nick:
							self.send_chan_msg(u'-err-', nchan, u'this buffer does not accept messages, only commands\n')
							return
					
					else:
						# private chat without the other user added yet;
						# pull in the other user
						utarget = None
						target = nchan.uchans[0].alias
						if target != from_nick:
							utarget = self.find_user(target)
							if utarget is None:
								self.send_chan_msg(u'-err-', nchan,
									u'\033[1;31mfailed to locate user "{0}"'.format(
										nchan.uchans[0].alias))
								return

							self.join_chan_obj(utarget, nchan, from_nick)
							self.start_logging(nchan)
							# fallthrough to send message

			now = time.time()
			msg = Message(nchan, now, from_nick, text)
			nchan.msgs.append(msg)
			nchan.latest = msg.ts

			if not from_nick.startswith(u'-') \
			and not from_nick == u'***':
				nchan.user_act_ts[from_nick] = now

			if len(nchan.msgs) > max_hist_mem:
				new_len = len(nchan.msgs) - msg_trunc_size
				print(' hist trunc:  [{0}] from {1} to {2}'.format(
					nchan.get_name(), len(nchan.msgs), new_len))
				while new_len > max_hist_mem:
					print('\033[1;31!!!\033[0m')
					new_len -= msg_trunc_size
				nchan.msgs = nchan.msgs[msg_trunc_size:]

			#self.refresh_chan(nchan)
			for uchan in nchan.uchans:
				if nchan.name is None or uchan.user.nick_re.search(text):
					#if len(nchan.uchans) == 1:
					#	break
					if PY2:
						if isinstance(uchan.user.nick, str):
							whoops('uchan.user.nick is bytestring')
						if isinstance(from_nick, str):
							whoops('from_nick is bytestring')

					if uchan.alias == 'r0c-status':
						if ping_self:
							uchan.last_ping = msg.sno
					else:
						if ping_self or uchan.user.nick != from_nick:
							uchan.last_ping = msg.sno
			
			if nchan not in self.dirty_ch:
				self.dirty_ch[nchan] = 1

			if nchan.log_fh:
				#print('logrotate counter at {0}'.format(nchan.log_ctr))
				if nchan.log_ctr >= self.messages_per_log_file:
					self.start_logging(nchan)
				
				nchan.log_ctr += 1
				nchan.log_fh.write((u' '.join(
					[hex(int(msg.ts*8.0))[2:].rstrip('L'),
					msg.user, msg.txt]) + u'\n').encode('utf-8'))


	def join_chan_obj(self, user, nchan, alias=None):
		with self.mutex:
			#print('{0} users in {1}, {2} messages; {3} is in {4} channels'.format(
			#	len(nchan.uchans), nchan.get_name(), len(nchan.msgs), user.nick, len(user.chans)))

			for uchan in user.chans:
				if uchan.nchan == nchan:
					return uchan
			
			self.num_joins += 1
			uchan = UChannel(user, nchan, alias)
			user.chans.append(uchan)
			nchan.uchans.append(uchan)
			nchan.user_act_ts[user.nick] = time.time()
			self.send_chan_msg(u'--', nchan,
				u'\033[1;32m{0}\033[22m has joined'.format(user.nick), False)
			uchan.last_read = nchan.msgs[-1].sno
			return uchan


	def get_pub_chan(self, name):
		for ch in self.pub_ch:
			if ch.name == name:
				return ch
		return None


	def get_priv_chan(self, user, alias):
		for ch in user.chans:
			if ch.alias == alias:
				return ch
		return None


	def join_pub_chan(self, user, name):
		with self.mutex:
			name = name.strip()
			nchan = self.get_pub_chan(name)
			if nchan is None:
				nchan = NChannel(name, u'#{0} - no topic has been set'.format(name))
				nchan.msgs.append(Message(nchan, time.time(), u'--', \
					u'\033[36mchannel created at \033[1m{0}'.format(
						datetime.datetime.utcnow().strftime('%Y-%m-%d, %H:%M:%SZ'))))
				#if nchan.name != 'xld':
				self.load_chat_log(nchan)
				#self.task_queue.put([self.load_chat_log, [nchan], {}])
				self.pub_ch.append(nchan)
			
			ret = self.join_chan_obj(user, nchan)
			user.new_active_chan = ret
			return ret


	def join_priv_chan(self, user, alias):
		with self.mutex:
			uchan = self.get_priv_chan(user, alias)
			if uchan is None:
				nchan = NChannel(None, u'DM with [[uch_a]]')
				self.priv_ch.append(nchan)
				uchan = self.join_chan_obj(user, nchan)
				uchan.alias = alias
			return uchan


	def broadcast_banner(self, msg):
		with self.mutex:
			chans = {}
			for user in self.users:
				if user.active_chan \
				and user.active_chan.nchan not in chans:
					chans[user.active_chan.nchan] = 1
			
			if not msg:
				for nchan in chans:
					if hasattr(nchan, 'topic_bak'):
						nchan.topic = nchan.topic_bak
						del nchan.topic_bak
				for user in self.users:
					if user.active_chan:
						user.client.refresh(False)
			else:
				for nchan in chans:
					if not hasattr(nchan, 'topic_bak'):
						nchan.topic_bak = nchan.topic
					nchan.topic = msg

				#print('broadcast: {0}'.format(msg))
				for user in self.users:
					if user.active_chan:
						#print('         : {0} ->'.format(user))
						to_send = u'\033[H{0}\033[K'.format(msg)
						user.client.screen[0] = to_send
						user.client.say(to_send.encode(
							user.client.codec, 'backslashreplace'))


	def broadcast_message(self, msg, severity=1):
		""" 1=append, 2=append+scroll """
		with self.mutex:
			for nchan in self.pub_ch:
				self.send_chan_msg(u'--', nchan, msg)
			
			for nchan in self.priv_ch:
				self.send_chan_msg(u'--', nchan, msg)
			
			if severity > 1:
				for user in self.users:
					if user.active_chan:
						if not user.active_chan.lock_to_bottom:
							user.active_chan.lock_to_bottom = True
							user.client.need_full_redraw = True
					else:
						user.client.say(u"\n [[ broadcast message ]]\n {0}\n".format(
							msg).replace(u"\n", u"\r\n").encode('utf-8'))


	def part_chan(self, uchan):
		with self.mutex:
			self.num_parts += 1
			user = uchan.user
			nchan = uchan.nchan
			i = None
			if user.active_chan == uchan:
				i = user.chans.index(uchan)
			user.chans.remove(uchan)
			nchan.uchans.remove(uchan)
			
			try: del nchan.user_act_ts[user.nick]
			except: pass

			if i:
				if len(user.chans) <= i:
					i -= 1
				user.new_active_chan = user.chans[i]
			
			self.send_chan_msg(u'--', nchan,
				u'\033[1;33m{0}\033[22m has left'.format(user.nick))

			if not nchan.uchans:
				print(' close chan:  [{0}]'.format(nchan.get_name()))

				if nchan.log_fh:
					nchan.log_fh.close()
				
				ch_list = self.pub_ch
				if not nchan.name:
					ch_list = self.priv_ch
				
				ch_list.remove(nchan)


	def load_chat_log(self, nchan):
		if not nchan:
			return

		#print('  chan hist:  scanning files')
		t1 = time.time()

		log_dir = u'{0}chan/{1}'.format(EP.log, nchan.name)
		try: os.makedirs(log_dir)
		except: pass

		if PY2:
			# os.walk stats all files (bad over nfs)
			files = os.listdir(log_dir.encode('utf-8'))
		else:
			files = []
			for (dirpath, dirnames, filenames) in os.walk(log_dir):
				files.extend(filenames)
				break
		
		re_chk = re.compile('^[0-9]{4}-[0-9]{4}-[0-9]{6}-*$')

		#total_size = 0
		#for fn in sorted(files):
		#	total_size += os.path.getsize(
		#		'{0}/{1}'.format(log_dir, fn))
		#
		#do_broadcast = (total_size > 1024*1024)
		#if do_broadcast:
		#	self.broadcast_banner('\033[1;37;45m [ LOADING CHATLOG ] \033[0;42m')
		#	# daily dose

		#print('  chan hist:  reading files')
		ln = u'???'
		t2 = time.time()
		chunks = [nchan.msgs]
		n_left = MAX_HIST_LOAD - len(nchan.msgs)
		bytes_loaded = 0
		try:
			for fn in reversed(sorted(files)):
				if not re_chk.match(fn):
					# unexpected file in log folder, skip it
					continue

				chunk = []
				with open('{0}/{1}'.format(log_dir, fn), 'rb') as f:
					f.readline()  # discard version info
					for ln in f:
						ts, user, txt = \
							ln.decode('utf-8').rstrip(u'\n').split(u' ', 2)

						chunk.append(Message(None, int(ts, 16)/8.0, user, txt))
						
					bytes_loaded += f.tell()

				#if chunk:
				#	chunk.append(Message(None, int(ts, 16)/8.0, '--', \
				#		'\033[36mend of log file "{0}"'.format(fn)))
				
				if len(chunk) > n_left:
					chunk = chunk[-n_left:]
				
				chunks.append(chunk)
				n_left -= len(chunk)
				if n_left <= 0:
					break
		except:
			whoops(ln)

		#print('  chan hist:  merging {0} chunks'.format(len(chunks)))
		nchan.msgs = []
		for chunk in reversed(chunks):
			#print('\nadding {0} messages:\n  {1}'.format(
			#	len(chunk), '\n  '.join(str(x) for x in chunk)))
			nchan.msgs.extend(chunk)

		#print('  chan hist:  setting {0} serials'.format(len(nchan.msgs)))
		for n, msg in enumerate(nchan.msgs):
			msg.sno = n

		t3 = time.time()
		print('  chan hist:  {0} msgs, {1:.0f} kB, {2:.2f} + {3:.2f} sec, #{5}'.format(
			MAX_HIST_LOAD - n_left, bytes_loaded/1024.0, t2-t1, t3-t2, t3-t1, nchan.name))
		
		#if do_broadcast:
		#	self.broadcast_banner(None)

		#if nchan.name != 'xst':
		self.start_logging(nchan, chunks[0])


	def start_logging(self, nchan, chat_backlog=None):
		if nchan.name is not None:
			log_dir = '{0}chan/{1}'.format(EP.log, nchan.name)
		else:
			log_dir = '{0}pm/{1}'.format(EP.log,
				'/'.join([x.user.nick for x in nchan.uchans]))

		if nchan.log_fh:
			nchan.log_fh.close()
		else:
			try: os.makedirs(log_dir)
			except: pass
		
		ts = datetime.datetime.utcnow().strftime('%Y-%m%d-%H%M%S')
		log_fn = '{0}/{1}'.format(log_dir, ts)

		while os.path.isfile(log_fn):
			log_fn += '-'

		nchan.log_ctr = 0
		nchan.log_fh = open(log_fn, 'wb')
		nchan.log_fh.write(u'1 {0:x}\n'.format(
			int(time.time())).encode('utf-8'))

		#print('opened log file {0}'.format(log_fn))

		if chat_backlog:
			#print('appending backlog ({0} messages)'.format(len(chat_backlog)))
			for msg in chat_backlog:
				nchan.log_ctr += 1
				nchan.log_fh.write((u' '.join(
					[hex(int(msg.ts*8.0))[2:].rstrip('L'),
					msg.user, msg.txt]) + u'\n').encode('utf-8'))
			
			# potential chance that a render goes through
			# before the async job processor kicks in
			self.dirty_ch[nchan] = 1
			for uchan in nchan.uchans:
				uchan.user.client.need_full_redraw = True




# -*- coding: utf-8 -*-
from __future__ import print_function
from .__init__ import *
if __name__ == '__main__':
  raise RuntimeError('\r\n{0}\r\n\r\n  this file is part of retr0chat.\r\n  enter the parent folder of this file and run:\r\n\r\n    python -m r0c <telnetPort> <netcatPort>\r\n\r\n{0}'.format('*'*72))

import gc

# make everything familiar to globals()
from .config  import *
from .util    import *
from .unrag   import *
from .ivt100  import *
from .inetcat import *
from .itelnet import *
from .chat    import *
from .user    import *
from .world   import *

if PY2:
  from Queue import Queue
else:
  from queue import Queue



def memory_dump():
	import gc
	import _pickle as cPickle
	with open("profiler-results.memory", 'wb') as dump:
		for obj in gc.get_objects():
			i = id(obj)
			size = sys.getsizeof(obj, 0)
			#    referrers = [id(o) for o in gc.get_referrers(obj) if hasattr(o, '__class__')]
			referents = [id(o) for o in gc.get_referents(obj) if hasattr(o, '__class__')]
			if hasattr(obj, '__class__'):
				cls = str(obj.__class__)
				cPickle.dump({'id': i, 'class': cls, 'size': size, 'referents': referents}, dump)



def get_obj_name(target_id):
	variables = {**locals(), **globals()}
	for var in variables:
		exec('var_id=id({0})'.format(var))
		if var_id == target_id:
			exec('found={0}'.format(var))
	if found:
		print(found)
		print(id(found))
	else:
		print('/')

def find_leaked_messages():
	n_hits = 0
	for obj in gc.get_objects():
		if not type(obj) is Message:
			continue
		n_hits += 1
		if n_hits > 1000:
			break
		
		ref_objs = gc.get_referents(obj)
		referents = [id(o) for o in ref_objs if hasattr(o, '__class__')]
		print('obj,ref:',id(obj),referents)
		print('obj:',obj)
		print('ref:',ref_objs)
		print()

repl_notepad = """

from .diag import *
find_leaked_messages()

n_hits = 0
for obj in gc.get_objects():
  if not hasattr(obj, '__class__'):
    continue
  if type(obj) is Message:
    n_hits += 1
    if n_hits > 10000:
      break
    i = id(obj)
    size = sys.getsizeof(obj, 0)
    name = obj.__name__ if hasattr(obj, '__name__') else '/'
    referents = [id(o) for o in gc.get_referents(obj) if hasattr(o, '__class__')]
    print('id: {0}, size: {1}, class: {2}, name: {3}, referents: {4}'.format(
      i, size, str(obj.__class__), name, ', '.join([str(x) for x in referents])))

"""





# dumping ground for mostly useless code below

def test_ansi_annotation():
	rangetype = range
	try: rangetype = xrange
	except: pass
	ansi_txt = '\033[1;33mHello \033[1;32mWorld\033[0m! This \033[7mis\033[0m a test.\033[A'
	ansi_txt = '\033[mf\033[s\033[w\033[has\033[3451431613gt\033[m \033[s\033[g\033[s\033[g\033[s\033[gcod\033[me\033[x'
	rv = visual_indices(ansi_txt)
	print(' '.join(ansi_txt.replace('\033', '*')))
	print(' '.join([str(x%10) for x in rangetype(len(ansi_txt))]))
	print(' '.join([str(x) for x in rv]))
	print('{0} {1}'.format(visual_length(ansi_txt), len(rv)))
	visual = ''
	for ofs in rv:
		visual += ansi_txt[ofs]
	print('[{0}]'.format(visual))

	for outer_n in rangetype(3):

		t0 = time.time()
		for n in rangetype(100000):
			rv = visual_indices(ansi_txt)
		print(str(time.time() - t0))
		
		t0 = time.time()
		for n in rangetype(100000):
			rv = visual_length(ansi_txt)
		print(str(time.time() - t0))




# -*- coding: utf-8 -*-
if __name__ == '__main__':
	raise RuntimeError('\r\n{0}\r\n\r\n  this file is part of retr0chat.\r\n  enter the parent folder of this file and run:\r\n\r\n    python -m r0c <telnetPort> <netcatPort>\r\n\r\n{0}'.format('*'*72))


# the admin password
ADMIN_PWD = u'ixbaka'

# salt for autogenerated nicknames based on IP
NICK_SALT = u'lammo/'


# two example values are listed for each config attribute;
# the first for debug purposes, the second for regular use

# log incoming traffic from clients
LOG_RX = True
#LOG_RX = False

# log outgoing traffic to clients
LOG_TX = True
LOG_TX = False

# show negotiations etc
DBG = True
DBG = False

# start monitoring threads on ctrl-c
THREADMON = True
THREADMON = False

# show all traffic from clients
HEXDUMP_IN = True
HEXDUMP_IN = False

# show all traffic to clients
HEXDUMP_OUT = True
HEXDUMP_OUT = False

# filter packets larger than N bytes from being hexdumped
HEXDUMP_TRUNC = 65535
HEXDUMP_TRUNC = 1
HEXDUMP_TRUNC = 128

# set true for a really bad "slow network" simulation
SLOW_MOTION_TX = True
SLOW_MOTION_TX = False

# force clients into linemode (to debug linemode UI)
FORCE_LINEMODE = True
FORCE_LINEMODE = False

# dump statistics every 2 seconds
BENCHMARK = True
BENCHMARK = False

# logrotate
MESSAGES_PER_LOG_FILE = 16
MESSAGES_PER_LOG_FILE = 512
MESSAGES_PER_LOG_FILE = 131072

# max number of messages to load from disk when joining a channel
MAX_HIST_LOAD = 64
MAX_HIST_LOAD = 65536

# max number of messages in channel scrollback before truncating
MAX_HIST_MEM = 80
MAX_HIST_MEM = 98304

# number of messages to remove every time the chan gets 2big
# (may cause a full redraw of some clients focused on chan)
MSG_TRUNC_SIZE = 8
MSG_TRUNC_SIZE = 16384

# width of the hexdump, in bytes per line
HEX_WIDTH = 16



# -*- coding: utf-8 -*-
from __future__ import print_function
from .__init__ import *
if __name__ == '__main__':
	raise RuntimeError('\r\n{0}\r\n\r\n  this file is part of retr0chat.\r\n  enter the parent folder of this file and run:\r\n\r\n    python -m r0c <telnetPort> <netcatPort>\r\n\r\n{0}'.format('*'*72))

from .util import *



# copied from  http://xxyxyz.org/line-breaking/
#
# license clarified per email:
#   BSD or MIT since the EU doesn't recognize Public Domain
#
# only change so far is replacing len() with visual_length()
#
# select which unragger to use at the bottom of this file



def unrag_1_linear(text, width):
	words = text.split()
	count = len(words)
	offsets = [0]
	for w in words:
		offsets.append(offsets[-1] + visual_length(w))

	minima = [0] + [10 ** 20] * count
	breaks = [0] * (count + 1)

	def cost(i, j):
		w = offsets[j] - offsets[i] + j - i - 1
		if w > width:
			return 10 ** 10 * (w - width)
		return minima[i] + (width - w) ** 2

	def smawk(rows, columns):
		stack = []
		i = 0
		while i < len(rows):
			if stack:
				c = columns[len(stack) - 1]
				if cost(stack[-1], c) < cost(rows[i], c):
					if len(stack) < len(columns):
						stack.append(rows[i])
					i += 1
				else:
					stack.pop()
			else:
				stack.append(rows[i])
				i += 1
		rows = stack

		if len(columns) > 1:
			smawk(rows, columns[1::2])

		i = j = 0
		while j < len(columns):
			if j + 1 < len(columns):
				end = breaks[columns[j + 1]]
			else:
				end = rows[-1]
			c = cost(rows[i], columns[j])
			if c < minima[columns[j]]:
				minima[columns[j]] = c
				breaks[columns[j]] = rows[i]
			if rows[i] < end:
				i += 1
			else:
				j += 2

	n = count + 1
	i = 0
	offset = 0
	while True:
		r = min(n, 2 ** (i + 1))
		edge = 2 ** i + offset
		smawk(range(0 + offset, edge), range(edge, r + offset))
		x = minima[r - 1 + offset]
		for j in range(2 ** i, r - 1):
			y = cost(j + offset, r - 1 + offset)
			if y <= x:
				n -= j
				i = 0
				offset += j
				break
		else:
			if r == n:
				break
			i = i + 1

	lines = []
	j = count
	while j > 0:
		i = breaks[j]
		lines.append(u' '.join(words[i:j]))
		j = i
	lines.reverse()
	return lines



from collections import deque 

def unrag_2_binary(text, width):
    words = text.split()
    count = len(words)
    offsets = [0]
    for w in words:
        offsets.append(offsets[-1] + visual_length(w))

    minima = [0] * (count + 1)
    breaks = [0] * (count + 1)

    def c(i, j):
        w = offsets[j] - offsets[i] + j - i - 1
        if w > width:
            return 10 ** 10 * (w - width)
        return minima[i] + (width - w) ** 2

    def h(l, k):
        low, high = l + 1, count
        while low < high:
            mid = (low + high) // 2
            if c(l, mid) <= c(k, mid):
                high = mid
            else:
                low = mid + 1
        if c(l, high) <= c(k, high):
            return high
        return l + 2

    q = deque([(0, 1)])
    for j in range(1, count + 1):
        l = q[0][0]
        if c(j - 1, j) <= c(l, j):
            minima[j] = c(j - 1, j)
            breaks[j] = j - 1
            q.clear()
            q.append((j - 1, j + 1))
        else:
            minima[j] = c(l, j)
            breaks[j] = l
            while c(j - 1, q[-1][1]) <= c(q[-1][0], q[-1][1]):
                q.pop()
            q.append((j - 1, h(j - 1, q[-1][0])))
            if j + 1 == q[1][1]:
                q.popleft()
            else:
                q[0] = q[0][0], (q[0][1] + 1)

    lines = []
    j = count
    while j > 0:
        i = breaks[j]
        lines.append(u' '.join(words[i:j]))
        j = i
    lines.reverse()
    return lines



def unrag_3_divide(text, width):
    words = text.split()
    count = len(words)
    offsets = [0]
    for w in words:
        offsets.append(offsets[-1] + visual_length(w))

    minima = [0] + [10 ** 20] * count
    breaks = [0] * (count + 1)

    def cost(i, j):
        w = offsets[j] - offsets[i] + j - i - 1
        if w > width:
            return 10 ** 10
        return minima[i] + (width - w) ** 2

    def search(i0, j0, i1, j1):
        stack = [(i0, j0, i1, j1)]
        while stack:
            i0, j0, i1, j1 = stack.pop()
            if j0 < j1:
                j = (j0 + j1) // 2
                for i in range(i0, i1):
                    c = cost(i, j)
                    if c <= minima[j]:
                        minima[j] = c
                        breaks[j] = i
                stack.append((breaks[j], j+1, i1, j1))
                stack.append((i0, j0, breaks[j]+1, j))

    n = count + 1
    i = 0
    offset = 0
    while True:
        r = min(n, 2 ** (i + 1))
        edge = 2 ** i + offset
        search(0 + offset, edge, edge, r + offset)
        x = minima[r - 1 + offset]
        for j in range(2 ** i, r - 1):
            y = cost(j + offset, r - 1 + offset)
            if y <= x:
                n -= j
                i = 0
                offset += j
                break
        else:
            if r == n:
                break
            i = i + 1

    lines = []
    j = count
    while j > 0:
        i = breaks[j]
        lines.append(u' '.join(words[i:j]))
        j = i
    lines.reverse()
    return lines



def unrag_4_shortest(text, width):
    words = text.split()
    count = len(words)
    offsets = [0]
    for w in words:
        offsets.append(offsets[-1] + visual_length(w))

    minima = [0] + [10 ** 20] * count
    breaks = [0] * (count + 1)
    for i in range(count):
        j = i + 1
        while j <= count:
            w = offsets[j] - offsets[i] + j - i - 1
            if w > width:
                break
            cost = minima[i] + (width - w) ** 2
            if cost < minima[j]:
                minima[j] = cost
                breaks[j] = i
            j += 1

    lines = []
    j = count
    while j > 0:
        i = breaks[j]
        lines.append(u' '.join(words[i:j]))
        j = i
    lines.reverse()
    return lines



def bench_unrag(fname):
	t0 = time.time()
	n_lines = 0
	trend = {}
	with open(fname, 'rb') as f:
		for ln in f:
			n_lines += 1

	with open(fname, 'rb') as f:
		for n, ln in enumerate(f):
			if n % 8192 == 8191:
				t = time.time()
				print('{0:7} / {1}  {2:.3f}%  {3:.0f} p/s'.format(
					n, n_lines, n*100.0/n_lines, n/(t-t0)))
			ln = ln.decode('utf-8')
			if len(ln) < 40:
				ln = ln + ' ' + ln
			w = unrag(ln, 40)
			try:
				trend[len(w)] += 1
			except:
				trend[len(w)] = 1

	for k, v in sorted(trend.items()):
		print('{0} lines: {1} occurences'.format(k, v))

	t = time.time()
	print(n_lines)
	print(t-t0)



def unrag_layout_test_dump():
	scrw = 572
	msg = "012345678901234567890 Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
	msg = msg + ' ' + msg
	wrappers = [ unrag_1_linear, unrag_2_binary, unrag_3_divide, unrag_4_shortest ]
	avail = scrw/len(wrappers)
	last_txt = None
	for msgw in range(10, avail):
		print(msgw)
		results = []
		for wrapper in wrappers:
			results.append(wrapper(' '.join(prewrap(msg, msgw)), msgw))
			#results[-1].insert(0, '{0} {1}'.format(msgw, wrapper.__name__))
		txt = u''
		for row in range(0, 1000):
			ln = u'\n'
			for nr, result in enumerate(results):
				if len(result) <= row:
					ln += u' ' * avail
				else:
					ln += '\033[1;3{0}m{1}{2}\033[1;30m.{3}'.format(
						nr+1, result[row],
						u' ' * (msgw-len(result[row])),
						u' ' * (avail-msgw-1))
			if not ln.strip():
				break
			txt += '\033[G' + ln
		if txt != last_txt:
			print(txt)
		last_txt = txt



def unrag_layout_test_interactive():
	try: input = raw_input
	except NameError: pass

	try:
		import msvcrt
	except:
		import sys, tty, termios
	def getch():
		try:
			return msvcrt.getch()
		except:
			fd = sys.stdin.fileno()
			old_cfg = termios.tcgetattr(fd)
			tty.setraw(sys.stdin.fileno())
			ret = sys.stdin.read(1)
			termios.tcsetattr(fd, termios.TCSADRAIN, old_cfg)
			return ret
	
	msg = "012345678901234567890 Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
	msg = msg + ' ' + msg
	wrappers = [ unrag_1_linear, unrag_2_binary, unrag_3_divide, unrag_4_shortest ]
	iwrapper = 0
	msgw = 20
	while True:
		if msgw < 3:
			msgw = 3
		if iwrapper < 0:
			iwrapper = len(wrappers) - 1
		if iwrapper >= len(wrappers):
			iwrapper = 0

		pwrap = ' '.join(prewrap(msg, msgw))
		wraps = []
		uniq_wraps = []
		for wrapper in wrappers:
			wrap = '\n'.join(wrapper(pwrap, msgw))
			wraps.append(wrap)
			if wrap not in uniq_wraps:
				uniq_wraps.append(wrap)

		wrapped = wraps[iwrapper]
		print('{0}\033[H\033[J\n  {1} {2}  ~{3}\n\n{4}'.format(
			'\n'*20, msgw, wrappers[iwrapper].__name__, len(uniq_wraps), wrapped))

		ch = getch()
		if ch == b'\x03': sys.exit(0)
		elif ch == b'a': iwrapper -= 1
		elif ch == b'd': iwrapper += 1
		elif ch == b'w': msgw -= 1
		elif ch == b's': msgw += 1



unrag_bench_results = """
unrag_1_linear
   8191 / 853993  0.959%  4315 p/s
  16383 / 853993  1.918%  4643 p/s
  24575 / 853993  2.878%  4771 p/s
  32767 / 853993  3.837%  4767 p/s
  40959 / 853993  4.796%  4801 p/s
  49151 / 853993  5.755%  4803 p/s
  57343 / 853993  6.715%  4841 p/s
0 lines: 22 occurences
1 lines: 66708 occurences
2 lines: 508953 occurences
3 lines: 88698 occurences
4 lines: 104985 occurences
5 lines: 81862 occurences
6 lines: 2764 occurences
7 lines: 1 occurences
166.26898884773254

unrag_2_binary
   8191 / 853993  0.959%  4710 p/s
  16383 / 853993  1.918%  5103 p/s
  24575 / 853993  2.878%  5248 p/s
  32767 / 853993  3.837%  5253 p/s
  40959 / 853993  4.796%  5289 p/s
  49151 / 853993  5.755%  5292 p/s
  57343 / 853993  6.715%  5332 p/s
0 lines: 22 occurences
1 lines: 36691 occurences
2 lines: 508243 occurences
3 lines: 93677 occurences
4 lines: 124877 occurences
5 lines: 87375 occurences
6 lines: 2996 occurences
7 lines: 72 occurences
8 lines: 16 occurences
9 lines: 21 occurences
10 lines: 3 occurences
149.93761730194092

unrag_3_divide
   8191 / 853993  0.959%  6596 p/s
  16383 / 853993  1.918%  7298 p/s
  24575 / 853993  2.878%  7580 p/s
  32767 / 853993  3.837%  7627 p/s
  40959 / 853993  4.796%  7713 p/s
  49151 / 853993  5.755%  7749 p/s
  57343 / 853993  6.715%  7816 p/s
0 lines: 22 occurences
1 lines: 68011 occurences
2 lines: 512204 occurences
3 lines: 84440 occurences
4 lines: 104785 occurences
5 lines: 81780 occurences
6 lines: 2750 occurences
7 lines: 1 occurences
101.88200402259827

unrag_4_shortest
   8191 / 853993  0.959%  8149 p/s
  16383 / 853993  1.918%  9167 p/s
  24575 / 853993  2.878%  9563 p/s
  32767 / 853993  3.837%  9664 p/s
  40959 / 853993  4.796%  9649 p/s
  49151 / 853993  5.755%  9495 p/s
  57343 / 853993  6.715%  9448 p/s
0 lines: 22 occurences
1 lines: 85310 occurences
2 lines: 497438 occurences
3 lines: 82328 occurences
4 lines: 104426 occurences
5 lines: 81719 occurences
6 lines: 2749 occurences
7 lines: 1 occurences
82.48605155944824
"""



# divide and shortest are the most A E S T H E T I C
# shortest wins just barely

#unrag = unrag_1_linear    # 166sec
#unrag = unrag_2_binary    # 150sec
#unrag = unrag_3_divide    # 102sec
unrag = unrag_4_shortest   #  82sec



# -*- coding: utf-8 -*-
from __future__ import print_function
from .__init__ import *
if __name__ == '__main__':
	raise RuntimeError('\r\n{0}\r\n\r\n  this file is part of retr0chat.\r\n  enter the parent folder of this file and run:\r\n\r\n    python -m r0c <telnetPort> <netcatPort>\r\n\r\n{0}'.format('*'*72))

import asyncore
import sys

from .util   import *
from .ivt100 import *



class NetcatServer(VT100_Server):

	def __init__(self, host, port, world, other_if):
		VT100_Server.__init__(self, host, port, world, other_if)
		self.user_config_path = EP.log + 'cfg.netcat'

	def gen_remote(self, socket, addr, user):
		return NetcatClient(self, socket, addr, self.world, user)



class NetcatClient(VT100_Client):
	
	def __init__(self, host, socket, address, world, user):
		VT100_Client.__init__(self, host, socket, address, world, user)
		
		self.looks_like_telnet = {
			b'\xff\xfe': 1,
			b'\xff\xfd': 1,
			b'\xff\xfc': 1,
			b'\xff\xfb': 1
		}
		# trick telnet into revealing itself:
		# request client status and location
		self.replies.put(b'\xff\xfd\x05\xff\xfd\x17')
		

	def handle_read(self):
		with self.world.mutex:
			if self.dead:
				print('XXX reading when dead')
				return

			data = self.recv(8192)
			if not data:
				if not self.dead:
					self.host.part(self)
				return
			
			if HEXDUMP_IN:
				hexdump(data, '-->>')

			if self.wire_log and LOG_RX:
				self.wire_log.write('{0:.0f}\n'.format(
					time.time()*1000).encode('utf-8'))
				hexdump(data, '>', self.wire_log)
			
			self.in_bytes += data

			if b'\xff' in data:
				ofs = 0
				while ofs >= 0:
					ofs = data.find(b'\xff', ofs)
					if ofs < 0:
						break
					if data[ofs:ofs+2] in self.looks_like_telnet:
						self.num_telnet_negotiations += 1
					ofs = ofs + 1
			
			try:
				src = u'{0}'.format(self.in_bytes.decode(self.codec))
				self.in_bytes = self.in_bytes[0:0]
			
			except UnicodeDecodeError as uee:
				if len(self.in_bytes) < uee.start + 6:
					print('need more data to parse unicode codepoint at {0} in {1}'.format(
						uee.start, len(self.in_bytes)))
					hexdump(self.in_bytes[-8:], 'XXX ')
					src = u'{0}'.format(self.in_bytes[:uee.start].decode(self.codec))
					self.in_bytes = self.in_bytes[uee.start:]
				else:
					# it can't be helped
					print('warning: unparseable data:')
					hexdump(self.in_bytes, 'XXX ')
					src = u'{0}'.format(self.in_bytes[:uee.start].decode(self.codec, 'backslashreplace'))
					self.in_bytes = self.in_bytes[0:0]  # todo: is this correct?
			
			self.in_text += src

			if self.wizard_stage is not None \
			and len(self.in_text_full) < 1024:
				self.in_text_full += src
			
			self.read_cb(False, len(src))




# -*- coding: utf-8 -*-
from __future__ import print_function
from .__init__ import *
if __name__ == '__main__':
	raise RuntimeError('\r\n{0}\r\n\r\n  this file is part of retr0chat.\r\n  enter the parent folder of this file and run:\r\n\r\n    python -m r0c <telnetPort> <netcatPort>\r\n\r\n{0}'.format('*'*72))

import traceback
import threading
import struct
import time
import sys
import os
import platform

from .config import *



print_mutex = threading.Lock()
if PY2:
	import __builtin__
	def print(*args, **kwargs):
		args = list(args)
		try:
			if WINDOWS and u'\033' in args[0]:
				args[0] = strip_ansi(args[0])
		except: pass
		
		with print_mutex:
			t = time.strftime('%H%M%S ')
			__builtin__.print(t + str(args[0] if args else u''
				).replace(u'\n', u'\n'+t), *args[1:], **kwargs)
else:
	import builtins
	def print(*args, **kwargs):
		args = list(args)
		try:
			if WINDOWS and u'\033' in args[0]:
				args[0] = strip_ansi(args[0])
		except: pass
		
		with print_mutex:
			t = time.strftime('%H%M%S ')
			builtins.print(t + str(args[0] if args else u''
				).replace(u'\n', u'\n'+t), *args[1:], **kwargs)

def num(c):
	try:
		return int(c)
	except:
		return None

def b2hex(data):
	if PY2:
		return ' '.join(map(lambda b: format(ord(b), "02x"), data))
	else:
		if type(data) is str:
			return ' '.join(map(lambda b: format(ord(b), "02x"), data))
		else:
			return ' '.join(map(lambda b: format(b, "02x"), data))

def hexdump(pk, prefix='', file=None):
	if file is not None:
		line_fmt = u'{0} {2}{3}{4}'
		hex_width = 4
		blk_width = 4
	else:
		line_fmt = u'{0}{1:8x}  {2}{3} {4}'
		hex_width = HEX_WIDTH
		blk_width = 8

	lpk = len(pk)
	ofs = 0
	hexofs = 0
	hexlen = 0
	hexstr = ''
	ascstr = ''
	ascstr_width = int(hex_width * 100 / 32.0 - 0.5)  # 32h = 100a, 16h = 50a
	while ofs < lpk:
		hexstr += b2hex(pk[ofs:ofs+blk_width])
		hexstr += ' '
		if PY2:
			ascstr += ''.join(map(lambda b: b if ord(b) >= 0x20 and ord(b) < 0x7f else '.', pk[ofs:ofs+blk_width]))
		else:
			ascstr += ''.join(map(lambda b: chr(b) if b >= 0x20 and b < 0x7f else '.', pk[ofs:ofs+blk_width]))
		hexlen += blk_width
		ofs += blk_width
		
		if hexlen >= hex_width or ofs >= lpk:
			txt = line_fmt.format(prefix, hexofs, hexstr,
				u' '*(ascstr_width-len(hexstr)), ascstr)
			
			if file is not None:
				file.write((txt + u'\n').encode('utf-8'))
			else:
				print(txt)

			hexofs = ofs
			hexstr = ''
			hexlen = 0
			ascstr = ''
		else:
			hexstr += ' '
			ascstr += ' '

def test_hexdump():
	try: from StringIO import StringIO as bio
	except: from io import BytesIO as bio
	
	v = b''
	for n in range(5):
		print()
		v += b'a'
		fobj = bio()
		hexdump(v, '>', fobj)
		print(fobj.getvalue().decode('utf-8').rstrip('\n') + '$')
		fobj.close()
	
	v = b''
	for n in range(18):
		print()
		v += b'a'
		hexdump(v, '>')
	
	sys.exit(0)



azAZ = u'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

def trunc(txt, maxlen):
	eoc = azAZ
	ret = u''
	clen = 0
	pend = None
	counting = True
	for input_ofs, ch in enumerate(txt):
		
		# escape sequences can never contain ESC;
		# treat pend as regular text if so
		if ch == u'\033' and pend:
			clen += len(pend)
			ret += pend
			counting = True
			pend = None
		
		if not counting:
			ret += ch
			if ch in eoc:
				counting = True
		else:
			if pend:
				pend += ch
				if pend.startswith(u'\033['):
					counting = False
				else:
					clen += len(pend)
					counting = True
				ret += pend
				pend = None
			else:
				if ch == u'\033':
					pend = u'{0}'.format(ch)
				else:
					ret += ch
					clen += 1
		
		if clen >= maxlen:
			return [ret, txt[input_ofs:]]
	
	return [ret, u'']



# adapted from trunc
def strip_ansi(txt):
	eoc = azAZ
	ret = u''
	pend = None
	counting = True
	for ch in txt:
		
		# escape sequences can never contain ESC;
		# treat pend as regular text if so
		if ch == u'\033' and pend:
			ret += pend
			counting = True
			pend = None
		
		if not counting:
			if ch in eoc:
				counting = True
		else:
			if pend:
				pend += ch
				if pend.startswith(u'\033['):
					counting = False
				else:
					ret += pend
					counting = True
				pend = None
			else:
				if ch == u'\033':
					pend = u'{0}'.format(ch)
				else:
					ret += ch
	return ret



# adapted from trunc
def visual_length(txt):
	eoc = azAZ
	clen = 0
	pend = None
	counting = True
	for ch in txt:
		
		# escape sequences can never contain ESC;
		# treat pend as regular text if so
		if ch == u'\033' and pend:
			clen += len(pend)
			counting = True
			pend = None
		
		if not counting:
			if ch in eoc:
				counting = True
		else:
			if pend:
				pend += ch
				if pend.startswith(u'\033['):
					counting = False
				else:
					clen += len(pend)
					counting = True
				pend = None
			else:
				if ch == u'\033':
					pend = u'{0}'.format(ch)
				else:
					clen += 1
	return clen



# 83% the speed of visual_length,
# good enough to stop maintaining it and swap w/ len(this)
def visual_indices(txt):
	eoc = azAZ
	ret = []
	pend_txt = None
	pend_ofs = []
	counting = True
	for n, ch in enumerate(txt):
		
		# escape sequences can never contain ESC;
		# treat pend as regular text if so
		if ch == u'\033' and pend_txt:
			ret.extend(pend_ofs)
			counting = True
			pend_txt = None
			pend_ofs = []
		
		if not counting:
			if ch in eoc:
				counting = True
		else:
			if pend_txt:
				pend_txt += ch
				pend_ofs.append(n)
				if pend_txt.startswith(u'\033['):
					counting = False
				else:
					ret.extend(pend_ofs)
					counting = True
				pend_txt = None
				pend_ofs = []
			else:
				if ch == u'\033':
					pend_txt = u'{0}'.format(ch)
					pend_ofs = [n]
				else:
					ret.append(n)
	return ret



def sanitize_ctl_codes(aside):
	plain = u''
	for pch in aside:
		nch = ord(pch)
		#print('read_cb inner  {0} / {1}'.format(b2hex(pch.encode('utf-8', 'backslashreplace')), nch))
		if nch < 0x20 and nch != 0x0b and nch != 0x0f:
			print('substituting non-printable \\x{0:02x}'.format(nch))
			plain += u'?'
		else:
			plain += pch
	return plain



FOREGROUNDS = {}
for luma, chars in enumerate([u'01234567',u'89abcdef']):
	for n, ch in enumerate(chars):
		FOREGROUNDS[ch] = u'\033[{0};3{1}'.format(luma, n)

BACKGROUNDS = {}
for n, ch in enumerate(u'01234567'):
	BACKGROUNDS[ch] = u';4{0}'.format(n)

def convert_color_codes(txt, preview=False):
	foregrounds = FOREGROUNDS
	backgrounds = BACKGROUNDS
	scan_from = 0
	while txt:
		ofs = txt.find(u'\x0b', scan_from)
		if ofs < 0:
			break
		
		scan_from = ofs + 1

		fg = None
		if len(txt) > ofs + 1:
			fg = txt[ofs+1]
		
		bg = None
		if len(txt) > ofs + 3 and txt[ofs+2] == u',':
			bg = txt[ofs+3]
		
		if fg in foregrounds:
			fg = foregrounds[fg]
		else:
			fg = None
			bg = None  # can't set bg without valid fg
		
		if bg in backgrounds:
			bg = backgrounds[bg]
		else:
			bg = None

		resume_txt = ofs + 1
		if fg:
			resume_txt += 1
			scan_from = len(fg) + 1
		if bg:
			resume_txt += 2
			scan_from += len(bg)

		preview_k = u''
		if preview:
			resume_txt = ofs + 1
			if fg:
				preview_k = u'K'

		if fg and bg:
			txt = u'{0}{1}{2}m{3}{4}'.format(
				txt[:ofs], fg, bg, preview_k, txt[resume_txt:])
		elif fg:
			txt = u'{0}{1}m{2}{3}'.format(
				txt[:ofs], fg, preview_k, txt[resume_txt:])
		else:
			txt = u'{0}K{1}'.format(
				txt[:ofs], txt[resume_txt:])

	scan_from = 0
	while txt:
		ofs = txt.find(u'\x0f', scan_from)
		if ofs < 0:
			break

		scan_from = ofs + 1
		txt = u'{0}\033[0m{2}{1}'.format(
			txt[:ofs], txt[scan_from:],
			u'O' if preview else u'')
	
	return txt



#B35_CHARS = tuple(u'0123456789abcdefghijkmnopqrstuvwxyz')
B35_CHARS = tuple(u'abcdefghijkmnopqrstuvwxyz')
B35_ATLAS = dict((c, i) for i, c in enumerate(B35_CHARS))
B35_BASE = len(B35_CHARS)
def b35enc(number):
	if not number:
		return B35_CHARS[0]

	prefix = u''
	if number < 0:
		prefix = u'-'
		number = abs(number)

	ret = u''
	while number:
		number, rem = divmod(number, B35_BASE)
		ret = B35_CHARS[rem] + ret

	return prefix + ret

def b35dec(b35str):
	factor = 1
	if b35str.startswith(u'-'):
		b35str = b35str[1:]
		factor = -1

	ret = 0
	for c in b35str:
		ret = ret * B35_BASE + B35_ATLAS[c]

	return factor * ret



def visualize_all_unicode_codepoints_as_utf8():
	stats = [0]*256
	nmax = sys.maxunicode + 1
	print('collecting all codepoints until {0}d, 0x{1:x}'.format(
		nmax, nmax))
	
	if PY2:
		to_unicode = unichr
		from_char = ord
	else:
		to_unicode = chr
		from_char = int
	
	for n in range(nmax):
		if n % 0x10000 == 0:
			print('at codepoint {0:6x} of {1:6x},  {2:5.2f}%'.format(
				n, nmax, (100.0 * n) / nmax))
		ch = to_unicode(n)
		
		try:
			bs = ch.encode('utf-8')
		except:
			# python2 allows encoding ud800 as \xed\xa0\x80 which is an illegal sequence in utf8;
			# python -c "for x in unichr(0xd800).encode('utf-8'): print '{0:2x}'.format(ord(x))"
			continue
		
		for b in bs:
			stats[from_char(b)] += 1

	print()
	for i, n in enumerate(stats):
		v = n
		if v == 0:
			v = 'illegal value'
		elif v == 1:
			v = 'single-use value'
		print('byte 0x{0:2x} occurences: {1}'.format(i, v))
	print()

#visualize_all_unicode_codepoints_as_utf8()



def prewrap(txt, maxlen):
	words = txt.split()
	ret = []
	for word in words:
		if len(word) < maxlen or visual_length(word) < maxlen:
			ret.append(word)
		else:
			while visual_length(word) >= maxlen:
				ret.append(word[:maxlen-1] + u'-')
				word = word[maxlen-1:]
			if word:
				ret.append(word)
	return ret



def whoops(extra=None):
	msg = """\
             __
   _      __/ /_  ____  ____  ____  _____
  | | /| / / __ \/ __ \/ __ \/ __ \/ ___/
  | |/ |/ / / / / /_/ / /_/ / /_/ (__  )
  |__/|__/_/ /_/\____/\____/ .___/____/
                          /_/"""
	exc = traceback.format_exc()
	if exc.startswith('None'):
		exc = ''.join(traceback.format_stack()[:-1])
	msg = '{0}\r\n{1}\r\n{2}</stack>'.format(
		msg, exc.rstrip(), '-'*64)
	print(msg)
	if extra:
		print('  {0}\n{1}\n'.format(extra, '-'*64))



def t_a_a_bt():
	ret = []
	for tid, stack in sys._current_frames().items():
		ret.append(u'\r\nThread {0} {1}'.format(tid, '='*64))
		for fn, lno, func, line in traceback.extract_stack(stack):
			ret.append(u'  File "{0}", line {1}, in {2}'.format(fn, lno, func))
			if line:
				ret.append(u'    {0}'.format(line.strip()))
	return u'\r\n'.join(ret)

thread_monitor_enabled = False
def monitor_threads():
	global thread_monitor_enabled
	if thread_monitor_enabled:
		return
	thread_monitor_enabled = True

	def stack_collector():
		while True:
			print('capturing stack')
			time.sleep(5)
			txt = t_a_a_bt()
			with open('r0c.stack', 'wb') as f:
				f.write(txt.encode('utf-8'))

	thr = threading.Thread(target=stack_collector, name='stk_col')
	thr.daemon = True
	thr.start()



def host_os():
	py_ver = '.'.join([str(x) for x in sys.version_info])
	ofs = py_ver.find('.final.')
	if ofs > 0:
		py_ver = py_ver[:ofs]

	bitness = struct.calcsize('P') * 8
	host_os = platform.system()
	return '{0} on {1}{2}'.format(py_ver, host_os, bitness)



def compat_chans_in_root():
	bad_dirs = []
	good_dirs = ['pm','chan','wire']
	for (dirpath, dirnames, filenames) in os.walk(EP.log):
		for d in dirnames:
			if d not in good_dirs:
				bad_dirs.append(d)
		break

	if bad_dirs:
		print()
		print('== performing upgrade in 5 seconds ==')
		print()
		print('Will move the following directories from [log] to [log/chan]:')
		print(', '.join(bad_dirs))
		print()
		print('PRESS CTRL-C TO ABORT')
		for n in range(5):
			print('{0} ...'.format(5-n))
			time.sleep(1)
	
		for d in bad_dirs:
			os.rename(
				'{0}{1}'.format(EP.log, d),
				'{0}chan/{1}'.format(EP.log, d))

		print('upgrade done \\o/')
		print()



# -*- coding: utf-8 -*-
from __future__ import print_function
from .__init__ import *
if __name__ == '__main__':
	raise RuntimeError('\r\n{0}\r\n\r\n  this file is part of retr0chat.\r\n  enter the parent folder of this file and run:\r\n\r\n    python -m r0c <telnetPort> <netcatPort>\r\n\r\n{0}'.format('*'*72))

import re
import hashlib
# debug imports
import code
import gc

from .util import *
from .chat import *



HELP_INTRO = u"""\
Useful commands:
   \033[36m/nick\033[0m  change your nickname
   \033[36m/help\033[0m  how-to and about

Text formatting:
  \033[36mCTRL-O\033[0m  reset text formatting
  \033[36mCTRL-B\033[0m  bold/bright text on/off
  \033[36mCTRL-K\033[0m  followed by a colour code:
       \033[36m2\033[0m  \033[32mgreen\033[0m,
    \033[36m15,4\033[0m  \033[1;37;44mbold white on blue\033[0m --
          say \033[1m/cmap\033[0m to see all options

Switching channels:
  \033[36mCTRL-E\033[0m  jump to active channel
  \033[36mCTRL-A\033[0m  jump to previous channel
  \033[36mCTRL-X\033[0m  jump to next channel
  \033[36m/3\033[0m      go to channel 3
  \033[36m/0\033[0m      go to this channel

Creating or joining the "general" chatroom:
  \033[36m/join #general\033[0m

Leaving a chatroom:
  \033[36m/part\033[0m

Changing your nickname:
  \033[36m/nick new_name\033[0m

Keybinds:
  \033[36mUp\033[0m / \033[36mDown\033[0m       input history
  \033[36mLeft\033[0m / \033[36mRight\033[0m    input field traversing
  \033[36mHome\033[0m / \033[36mEnd\033[0m      input field jump
  \033[36mPgUp\033[0m / \033[36mPgDown\033[0m   chatlog scrolling... \033[1mtry it :-)\033[0m

if you are using a mac, PgUp is fn-Shift-PgUp
"""



class User(object):
	def __init__(self, world, address):
		self.world = world
		self.admin = False           # set true after challenge success
		self.client = None           # the client which this object belongs to
		self.chans = []              # UChannel instances
		self.active_chan = None      # UChannel
		self.new_active_chan = None  # set for channel change
		self.old_active_chan = None  # last focused channel
		self.nick = None             # str
		self.nick_re = None          # regex object for ping assert
		self.nick_len = None         # visible segment for self

	def __unicode__(self):
		return u'User {0} {1}'.format(self.nick, self.client.addr[0])

	def __str__(self):
		return 'User {0} {1}'.format(self.nick, self.client.addr[0])

	def __repr__(self):
		return 'User({0}, {1})'.format(repr(self.nick), repr(self.client.addr[0]))

	def __lt__(self, other):
		return self.nick < other.nick

	def pattern_gen(self, depth=0):
		charset = u'/!@#$%^&*()_+-=[]{};:<>,.'
		for ch in charset:
			yield ch

		if depth < 2:  # <= 3 chars
			for ch1 in charset:
				for ch2 in self.pattern_gen(depth+1):
					yield ch1 + ch2

	def set_rand_nick(self):
		plain_base = NICK_SALT + u'{0}'.format(self.client.addr[0])
		for suffix in self.pattern_gen():
			plain = plain_base + suffix
			nv = hashlib.sha256(plain.encode('utf-8')).digest()
			if PY2:
				nv = int(nv.encode('hex'), 16)
			else:
				nv = int.from_bytes(nv, 'big')
			
			nv = b35enc(nv)[:6]

			if not self.world.find_user(nv):
				self.set_nick(nv)
				break



	def create_channels(self):
		# while true; do tail -n +3 ansi | iconv -t 'cp437//IGNORE' | iconv -f cp437 | while IFS= read -r x; do printf "$x\n"; done | sed -r "s/$/$(printf '\033[K')/"; printf '\033[J'; sleep 0.2; printf '\033[H'; done
		
		if self.client.codec in ['utf-8','cp437','shift_jis']:
			
			# the simple version
			text = u"""\
`1;30m________ ___ ________
`1;30m░▒▓█▀▀▀▀`37m █▀█ `30m▀▀▀▀█▓▒░   `0;36m┌pad1[`0mretr0chat r0c_ver`36m]pad2┐
`1;30m ░▒▓`36m █▀█ █ █ █▀▀ `30m▓▒░    `0;36m│`0mgithub.com/9001/r0c`36m│
`1;30m  ░▒`34m █   █▄█ █▄▄ `30m▒░     `0;36m╘═══════════════════╛
                             `34m  b. r0c_build `0m
"""
			# the messy version
			text = u"""\
`1;30m________ `37m__`36m_ `30m________
`1;30m░▒▓█▀▀▀▀`37m █▀`46m▓`0;1;30m ▀▀▀▀█▓▒░   `0;36m┌pad1[`0mret`1mr0c`22mhat r0c_ver`36m]pad2┐
`1;30m ░▒▓ `34;46m▒`0;1;36m▀█ `37;46m▓`0m `1;37;46m▓`0m `1;36m█▀`34m▀ `30m▓▒░    `0;36m│`0mgithub.com/9001/r0c`36m│
`1;30m  ░▒ `34m█   `36m█▄█ `34;46m▒`0;1;34m▄▄ `30m▒░     `0;36m╘═══════════════════╛
                             `34m  b. r0c_build `0m
"""
			
		else:
			# the simple version
			text = u"""
  `1;37m     /^\\           `0mretr0chat r0c_ver `36m-----
  `1;36m/^^  | |  /^^      `0mgithub.com/9001/r0c
  `1;34m|    \\_/  \\__      `0;36m------b. r0c_build `0m
"""

			# the messy version
			text = u"""`1;30m______    `37m_`30m    ______
`1;30m\\\\\\\\\\\\\\  `37m/ \\  `30m///////   `0mret`1mr0c`22mhat r0c_ver `36m-----
 `1;30m\\\\ `36m/`37m^^  | |  `36m/^`0;36m^`1;30m //    `0mgithub.com/9001/r0c
  `1;30m\\ `0;36m|    `1m\\_/  `0;36m\\__ `1;30m/     `0;36m------b. r0c_build `0m
"""

		pad1 = pad2 = u'──'
		if len(S_VERSION) > 3: pad1 = pad1[1:]
		if len(S_VERSION) > 4: pad2 = pad2[1:]
		if len(S_VERSION) > 5: pad1 = pad1[1:]
		if len(S_VERSION) > 6: pad2 = pad2[1:]

		text = text.replace(u'`', u'\033[').\
			replace('r0c_build', S_BUILD_DT).\
			replace('r0c_ver', S_VERSION).\
			replace('pad1', pad1).\
			replace('pad2', pad2)
		text += HELP_INTRO

		uchan = self.world.join_priv_chan(self, u'r0c-status')
		nchan = uchan.nchan
		nchan.topic = u'r0c readme (and status info)'

		msg = Message(nchan, time.time(), u'-nfo-', text)
		nchan.msgs.append(msg)

		if False:
			text = []
			lipsum1 = u"Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
			lipsum2 = u"Lorem ipsum dolor sit amet, \033[1;31mconsectetur\033[0m adipiscing elit, sed do eiusmod tempor incididunt ut \033[1;32mlabore et dolore magna\033[0m aliqua. Ut enim ad minim veniam, quis nostrud \033[1;33mexercitation ullamco laboris nisi ut aliquip ex ea\033[0m commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est labo\033[1;35mrum."
			for n in range(10):
				text.append(lipsum1)
				text.append(lipsum2)
			for ln in text:
				nchan.msgs.append(Message(nchan, time.time(), u'-nfo-', ln))

		self.new_active_chan = uchan
		


		if False:
			uchan = self.world.join_pub_chan(self, u'general')
			nchan = uchan.nchan
			if len(nchan.msgs) < 100:
				for n in range(1,200):
					txt = u'{1}_{0:03}     \\\\\\\\'.format(n,
						u'_{0:03}     \\\\\\\\\n'.format(n).join(
							str(v).rjust(v+4, u' ') for v in range(0, 12)))
					self.world.send_chan_msg(self.nick, nchan, txt)

		if False:
			uchan = self.world.join_pub_chan(self, u'smalltalk')
			nchan = uchan.nchan
			for n in range(1,3):
				txt = u'  message {0}\n      mes {0}'.format(n)
				self.world.send_chan_msg(self.nick, nchan, txt)

		self.client.handshake_world = True



	def admin_test(self, cmd, arg):
		if self.admin:
			return True

		print('denied exec:  /{0} {1} from {2} ({3})'.format(
			cmd, arg, self.nick, self.client.addr))

		inf = self.world.get_priv_chan(self, u'r0c-status').nchan
		
		self.world.send_chan_msg(u'-err-', inf, u"""\033[1;31m[denied]\033[0m
  don't move, the police are on the way
""")

		return False


	def exec_cmd(self, cmd_str):
		#print('handle {0}'.format(cmd_str))
		inf = self.world.get_priv_chan(self, u'r0c-status').nchan
		cmd = cmd_str # the command keyword
		arg = None    # single argument with spaces
		arg1 = None   # 1st of 2 arguments
		arg2 = None   # 2nd of 2 arguments
		
		ofs = cmd.find(u' ')
		if ofs > 0:
			cmd = cmd_str[:ofs]
			arg = cmd_str[ofs+1:]
		cmd = cmd.lower()

		if arg:
			arg1 = arg
			ofs = arg.find(u' ')
			if ofs > 0:
				arg1 = arg[:ofs].lower()
				arg2 = arg[ofs+1:]

		if cmd == u'help':
			self.help(arg, inf)

		elif cmd == u'me':
			self.world.send_chan_msg(u'***', self.active_chan.nchan,
				u'\033[1m{0}\033[22m {1}'.format(self.nick, arg))

		elif cmd == u'auth':
			if arg == self.world.core.password:
				self.admin = True
				self.world.send_chan_msg(u'-nfo-', inf, u"please don't break anything")
			else:
				self.world.send_chan_msg(u'-err-', inf, u'wrong password')

		elif cmd == u'nick' or cmd == u'n':
			if not arg:
				self.world.send_chan_msg(u'-err-', inf, u"""[invalid argument]
  usage:     /nick  new_nickname
  example:   /nick  spartacus
""")
				return

			# TODO: make this more lenient?
			legit_chars = azAZ
			legit_chars += u'0123456789_-'
			new_nick = u''
			nick_re = u''
			for ch in arg:
				if ch in legit_chars:
					new_nick += ch

			if not new_nick:
				self.world.send_chan_msg(u'-err-', inf, u"[invalid argument]\n  " +
					u"yooo EXCLUSIVELY illegal chars in new nick\n")
				return

			if new_nick != arg:
				self.world.send_chan_msg(u'-err-', inf, u"[invalid argument]\n  " +
					u"some illegal characters were removed\n")
				return

			if new_nick.startswith(u'-'):
				self.world.send_chan_msg(u'-err-', inf, u"[invalid argument]\n  " +
					u"nicks cannot start with \"-\" (dash)\n")
				return

			if len(new_nick) > 32:
				self.world.send_chan_msg(u'-err-', inf, u"[invalid argument]\n  " +
					u"too long\n")
				return

			other_user = None
			with self.world.mutex:
				if self.world.find_user(new_nick):
					self.world.send_chan_msg(u'-err-', inf, u"[invalid argument]\n  " +
						u"that nick is taken\n")
					return

				print('nick change:  {2} {0} -> {1}'.format(
					self.nick, new_nick, self.client.addr[0]))
				
				for uchan in self.chans:
					self.world.send_chan_msg(u'--', uchan.nchan,
						u'\033[1;36m{0}\033[22m changed nick to \033[1m{1}'.format(
							self.nick, new_nick), False)

				# update last-spoke tables
				now = time.time()
				for nchan in [x.nchan for x in self.chans]:
					nchan.user_act_ts[new_nick] = now
					try: del nchan.user_act_ts[self.nick]
					except: pass

				# update title in DM windows
				for nchan in self.world.priv_ch:
					for usr in nchan.uchans:
						if usr.alias == self.nick:
							usr.alias = new_nick

				self.set_nick(new_nick)



		elif cmd == u'topic' or cmd == u't':
			if not arg:
				self.world.send_chan_msg(u'-err-', inf, u"""[invalid argument]
  usage:     /topic  the_new_topic
  example:   /topic  cooking recipes
""")
				return

			uchan = self.active_chan
			nchan = uchan.nchan
			if nchan in self.world.priv_ch:
				self.world.send_chan_msg(u'-err-', inf, u"""[error]
  cannot change the topic of private channels
""")
				return

			old_topic = nchan.topic
			nchan.topic = arg
			self.world.send_chan_msg(u'--', nchan,
				u'\033[36m{0} has changed the topic from [\033[0m{1}\033[36m] -to-> [\033[0m{2}\033[36m]\033[0m'.format(
				self.nick, old_topic, arg))



		elif cmd == u'join' or cmd == u'j':
			if not arg or len(arg) < 2:
				self.world.send_chan_msg(u'-err-', inf, u"""[invalid arguments]
  usage:     /join  #channel_name
  example:   /join  #general
""")
				return
			
			if not arg.startswith(u'#'):
				self.world.send_chan_msg(u'-err-', inf, u"""[error]
  illegal channel name:  {0}
  channel names must start with #
""".format(arg))
				return

			nchan = self.world.join_pub_chan(self, arg[1:]).nchan
			# this is in charge of activating the new channel,
			# rapid part/join will crash us without this
			self.client.refresh(False)
			
			if False:
				# measure performance on chans with too many messages
				if len(self.active_chan.nchan.msgs) < 1048576:
					for n in range(0,1048576):
						if n % 16384 == 0:
							print(n)
						self.world.send_chan_msg(u'--', self.active_chan.nchan,
							u'large history load test {0}'.format(n))



		elif cmd == u'part' or cmd == u'p':
			if self.active_chan.alias == u'r0c-status':
				self.world.send_chan_msg(u'-err-', inf, u"""[error]
  cannot part the status channel
""".format(arg))
				return

			self.world.part_chan(self.active_chan)
			# this is in charge of activating the new channel,
			# rapid part/join will crash us without this
			self.client.refresh(False)



		elif cmd.isdigit():
			nch = int(cmd)
			if nch >= len(self.chans):
				self.world.send_chan_msg(u'-err-', inf, u"""[error]
  you only have {0} channels my dude
""".format(len(self.chans)))
				return

			self.new_active_chan = self.chans[nch]
			self.client.refresh(False)



		elif cmd == u'msg' or cmd == u'm':
			if not arg1 or not arg2:
				self.world.send_chan_msg(u'-err-', inf, u"""[invalid arguments]
  usage:     /msg   nickname   your message text
  example:   /msg   ed   hello world
""")
				return

			if not self.world.find_user(arg1):
				self.world.send_chan_msg(u'-err-', inf, u"""[user not found]
  "{0}" is not online
""".format(arg1))
				return

			uchan = self.world.join_priv_chan(self, arg1)
			self.new_active_chan = uchan
			self.world.send_chan_msg(self.nick, uchan.nchan, arg2)
			self.client.refresh(False)



		elif cmd == u'up' or cmd == u'u':
			self.client.scroll_cmd = -(self.client.h - 4)
		
		elif cmd == u'down' or cmd == u'd':
			self.client.scroll_cmd = +(self.client.h - 4)
		
		elif cmd == u'latest' or cmd == u'l':
			self.active_chan.lock_to_bottom = True
			self.client.need_full_redraw = True
			self.client.refresh(False)

		elif cmd == u'redraw' or cmd == u'r':
			if self.client.request_terminal_size('redraw'):
				# returns true if event was scheduled for later
				return

			self.client.need_full_redraw = True
			self.client.refresh(False)



		elif cmd == u'fill':
			if not self.admin_test(cmd, arg):
				return

			for n in range(int(arg1)):
				self.world.send_chan_msg(
					self.nick, self.active_chan.nchan,
					u'{0} {1}'.format(arg2, n))



		elif cmd == u'names' or cmd == u'na':
			self.world.send_chan_msg(u'--', inf, u"{1} users in {0}: {2}".format(
				self.active_chan.nchan.get_name(),
				len(self.active_chan.nchan.uchans),
				u', '.join(sorted([x.user.nick for x in
					self.active_chan.nchan.uchans]))))



		elif cmd == u'status' or cmd == u'st':
			n_wizard = sum(1 for x in self.world.users if not x.active_chan)
			n_users = len(self.world.users) - n_wizard
			n_pub = len(self.world.pub_ch)
			n_priv = len(self.world.priv_ch) - n_users
			
			n_in_chans = 0
			seen_users = {}

			for chan in self.world.pub_ch:
				for user in [x.user for x in chan.uchans]:
					if user not in seen_users:
						seen_users[user] = 1
						n_in_chans += 1
			
			for chan in self.world.priv_ch:
				if len(chan.uchans) == 1:
					continue
				for user in [x.user for x in chan.uchans]:
					if user not in seen_users:
						seen_users[user] = 1
						n_in_chans += 1

			self.world.send_chan_msg(u'--', inf,
				u"{0} users + {1} in wizard, {2} in chans, {3} public + {4} private chans".format(
					n_users, n_wizard, n_in_chans, n_pub, n_priv))

			if self.admin:
				self.world.send_chan_msg(u'--', inf, u'----- users -----')
				for user in sorted(self.world.users):
					self.world.send_chan_msg(
						u'--', inf, u'{0} {1} {2}'.format(
						user.client.addr[0].ljust(15),
						u'ok ' if user.active_chan else u'wiz',
						user.nick))
				self.world.send_chan_msg(u'--', inf, u'----- chans -----')
				for chan in sorted(self.world.pub_ch):
					self.world.send_chan_msg(
						u'--', inf, u'{0}: {1}'.format(chan.name,
						u', '.join(sorted([x.user.nick for x in chan.uchans]))))
				self.world.send_chan_msg(u'--', inf, u'-----------------')



		elif cmd == u'a':
			activity = {}
			for uchan in self.chans:
				if uchan.hilights and uchan != self.active_chan:
					activity[uchan.last_ping] = uchan
			for uchan in self.chans:
				if uchan.activity and uchan != self.active_chan:
					activity[uchan.nchan.msgs[-1].ts] = uchan
			
			if activity:
				x, uchan = sorted(activity.items())[0]
				self.new_active_chan = uchan
				nchan = uchan.nchan
				for msg in nchan.msgs:
					if msg.sno > uchan.last_read:
						#print('1st unread msg ({0} > {1}) = {2}'.format(
						#	msg.sno, uchan.last_read, msg))
						jump_to = nchan.msgs.index(msg) - 5
						if jump_to < 0:
							jump_to = 0
						uchan.jump_to_msg(jump_to)
						break
				#print('jumping to activity in {0}'.format(
				#	self.new_active_chan.nchan.get_name()))
			
			elif self.old_active_chan:
				self.new_active_chan = self.old_active_chan
				#print('jumping to last active, {0}'.format(
				#	self.new_active_chan.nchan.get_name()))
			else:
				print('cannot jump, no hilights or prev chan')
			
			self.client.need_full_redraw = True
			self.client.refresh(False)



		elif cmd == u'goto' or cmd == u'g':
			ch = self.active_chan
			nch = self.active_chan.nchan
			if not arg:
				self.world.send_chan_msg(u'--', inf, u"""[goto]
  {1} msgs since {2} in {0}
  
  command usage:
    /g 19:47             jump to time
    /g 2018-01-21        jump to date
    /g 2018-01-21 19:47  jump to datetime
    /g 3172              jump to message
    /g 34%               jump to offset
    /l                   jump to most recent
""".format(
				nch.get_name(),
				len(nch.msgs),
				datetime.datetime.utcfromtimestamp(nch.msgs[0].ts).strftime('%Y-%m-%d, %H:%M')))

			else:
				tfmt = '%Y-%m-%dT%H:%M:%S'

				m = re.match('(^[0-9]+)$', arg)
				if m:
					ch.jump_to_msg(int(m.group(1)))
					return
				
				m = re.match('(^[0-9\.]+)%$', arg)
				if m:
					ch.jump_to_msg(int(float(m.group(1))*len(nch.msgs)/100.0))
					return

				m = re.match('(^[0-9]{4}-[0-9]{2}-[0-9]{2}) ([0-9]{2}:[0-9]{2})$', arg)
				if m:
					ht = u'{0}T{1}:00'.format(*m.groups())
					ch.jump_to_time(datetime.datetime.strptime(ht, tfmt))
					return

				m = re.match('(^[0-9]{4}-[0-9]{2}-[0-9]{2})$', arg)
				if m:
					ht = u'{0}T00:00:00'.format(m.group(1))
					ch.jump_to_time(datetime.datetime.strptime(ht, tfmt))
					return

				m = re.match('(^[0-9]{2}:[0-9]{2})$', arg)
				if m:
					ht = u'{0}T{1}:00'.format(time.strftime('%Y-%m-%d'), m.group(1))
					ch.jump_to_time(datetime.datetime.strptime(ht, tfmt))
					return

				self.world.send_chan_msg(u'-err-', inf, u"""[goto]
  invalid argument format, see /g for help
""")



		elif cmd == u'search' or cmd == u'srch' or cmd == u's':
			ch = self.active_chan
			nch = self.active_chan.nchan
			if not arg:
				self.world.send_chan_msg(u'--', inf, u"""[search]
  plaintext search:
    /s fore       finds messages like "before"

  regex search:
    /s s/\\bfore/  finds messages like "foremost"
""")



		elif cmd == u'sw':
			try: arg = int(arg)
			except: arg = None
			
			if not arg:
				self.world.send_chan_msg(u'-err-', inf, u"""[invalid arguments]
  usage:     /sw  your_screen_width
  example:   /sw  80
""")
				return

			self.client.w = arg
			self.world.send_chan_msg(u'-nfo-', inf, \
				u'screen width: {0} letters'.format(self.client.w), False)



		elif cmd == u'sh':
			try: arg = int(arg)
			except: arg = None
			
			if not arg:
				self.world.send_chan_msg(u'-err-', inf, u"""[invalid arguments]
  usage:     /sh  your_screen_height
  example:   /sh  24
""")
				return

			self.client.h = arg
			self.world.send_chan_msg(u'-nfo-', inf, \
				u'screen height: {0} letters'.format(self.client.h), False)



		elif cmd == u'ss':
			if arg == u'0':
				arg = '100%'

			try: int_arg = int(arg)
			except: int_arg = None

			perc_arg = None
			if arg and arg.endswith(u'%'):
				try: perc_arg = int(arg[:-1])
				except: pass

			if int_arg is not None:
				if int_arg > 200:
					self.world.send_chan_msg(u'-err-', inf, u'whoa dude')
					return

				self.client.scroll_f = None
				self.client.scroll_i = int_arg
				self.world.send_chan_msg(u'-nfo-', inf, \
					u'scroll size: {0} lines'.format(self.client.scroll_i), False)
			
			elif perc_arg is not None:
				if perc_arg > 200:
					self.world.send_chan_msg(u'-err-', inf, u'whoa dude')
					return

				self.client.scroll_i = None
				self.client.scroll_f = perc_arg / 100.0
				self.world.send_chan_msg(u'-nfo-', inf, \
					u'scroll size: {0}% of screen'.format(self.client.scroll_f*100), False)

			else:
				self.world.send_chan_msg(u'-err-', inf, u"""[invalid arguments]
  usage:     /ss  lines_scrolled_per_pgup_pgdn
  example:   /sh  0     (entire screen)
  example:   /sh  10    (10 lines)
  example:   /sh  50%   (half the screen)
""")
			return



		elif cmd == u'by':
			self.client.bell = True
			self.world.send_chan_msg(u'--', inf,
				u'Audible alerts enabled. Disable with /bn', False)

		elif cmd == u'bn':
			self.client.bell = False
			self.world.send_chan_msg(u'--', inf,
				u'Audible alerts disabled. Enable with /by', False)



		elif cmd == u'cmap':
			msg = u"All foreground colours (0 to f) on default background,\n"
			msg += u"each code wrapped in [brackets] for readability:\n  "
			for n in range(0, 16):
				if n == 8:
					msg += u'\n  \033[1;3{0}m[{1:x}], '.format(n%8, n)
				else:
					msg += u'\033[3{0}m[{1:x}], '.format(n%8, n)
			
			msg += u"\033[0m\n\nEach background with black text:\n  \033[30m"
			for n in range(0, 8):
				msg += u'\033[4{0}m 0,{0} '.format(n)
			
			msg += u"\033[0m\n\nEach background with gray text:\n  \033[37m"
			for n in range(0, 8):
				msg += u'\033[4{0}m 7,{0} '.format(n)
			
			msg += u"\033[0m\n\nEach background with white text:\n  \033[1;37m"
			for n in range(0, 8):
				msg += u'\033[4{0}m f,{0} '.format(n)
			
			msg += u"\033[0m\n"
			self.world.send_chan_msg(u'-nfo-', inf, msg)



		elif cmd == u'sd':
			if not self.admin_test(cmd, arg):
				return
			
			msg = u"\033[31mserver shutdown requested by \033[1m{0}".format(self.nick)
			self.world.broadcast_message(msg, 2)
			
			def delayed_shutdown():
				time.sleep(0.5)
				self.world.core.shutdown()
			
			thr = threading.Thread(target=delayed_shutdown, name='shutd')
			thr.daemon = True
			thr.start()

		elif cmd == u'mem':
			if not self.admin_test(cmd, arg):
				return
			
			print('memdump started')
			memory_dump()
			print('memdump done')

		elif cmd == u'repl':
			if not self.admin_test(cmd, arg):
				return
			
			print('entering repl')
			#code.interact(locals=locals())
			code.InteractiveConsole(locals=globals()).interact()
			print('left repl')

		elif cmd == u'gc':
			if not self.admin_test(cmd, arg):
				return
			
			gc.collect()

		elif cmd == u'quit' or cmd == u'q' or cmd == u'exit':
			self.client.host.part(self.client)



		else:
			self.world.send_chan_msg(u'-err-', inf, u"""invalid command:  /{0}
  if you meant to send that as a message,
  escape the leading "/" by adding another "/"
""".format(cmd_str))





	def set_nick(self, new_nick):
		nick_re = u''
		# re.IGNORECASE doesn't work
		# this is dumb
		for ch in re.escape(new_nick):
			if not ch in azAZ:
				nick_re += ch
			else:
				nick_re += u'[{0}{1}]'.format(ch.lower(), ch.upper())
		
		self.nick = new_nick
		self.nick_re = re.compile(
			'(^|[^a-zA-Z0-9]){0}([^a-zA-Z0-9]|$)'.format(
				nick_re))

		self.nick_len = len(new_nick)
		if self.nick_len > self.client.w * 0.25:
			self.nick_len = int(self.client.w * 0.25)

		if self.active_chan:
			self.client.save_config()



	def help(self, arg, inf):
		if not arg:
			arg = u'topics'

		txt = None
		if arg == u'intro':
			txt = HELP_INTRO
		else:
			legit_chars = azAZ
			page = ''
			for ch in arg:
				if ch in legit_chars:
					page += ch
			try:
				with open(EP.doc + 'help-{0}.md'.format(page), 'rb') as f:
					txt = f.read().decode('utf-8')
			except:
				self.world.send_chan_msg(u'-err-', inf, u'that help page does not exist')
				return
		
		txt = txt.replace(u'\r', u'')
		txt = txt.replace(u'  \n', u'\n')

		txt = u'\033[0;30;46m{0}\033[K\n\033[0m\n'.format(u'=' * 32) + txt

		txt = txt.replace(u'\n\n| | |\n|-|-|\n', u'\n')
		if u'\n| ' in txt:
			txt = txt.replace(u'\n| ', u'\n').replace(u' | ', u'  ')
			if txt.startswith(u'| '):
				txt = txt[2:]

		txt = re.sub(r'\*\*`?([^\*`]+)`?\*\*', '\033[1;36m\\1\033[0m', txt)
		txt = re.sub(r'`([^`]+)`',             '\033[1;35m\\1\033[0m', txt)
		txt = re.sub(r'\n# ([^\n]*)\n',        '\n\033[1;33m=== \\1 ===\033[0m\n', txt)
		txt = txt.replace(u'\n', u'\r\n')

		self.world.send_chan_msg(u'-nfo-', inf, txt)



# -*- coding: utf-8 -*-
from __future__ import print_function
from .__init__ import *
if __name__ == '__main__':
	raise RuntimeError('\r\n{0}\r\n\r\n  this file is part of retr0chat.\r\n  enter the parent folder of this file and run:\r\n\r\n    python -m r0c <telnetPort> <netcatPort>\r\n\r\n{0}'.format('*'*72))

import threading
import datetime
import calendar

from .util import *



class NChannel(object):
	def __init__(self, name, topic):
		self.uchans = []       # UChannel instances
		self.msgs = []         # messages
		self.name = name
		self.topic = topic
		self.user_act_ts = {}  # str(nick) -> ts(last activity)

		self.log_fh = None     # active log file
		self.log_ctr = 0       # number of messages in file
	
	def get_name(self):
		if self.name:
			return u'#' + self.name
		ret = u', '.join(x.alias for x in self.uchans[:2])
		return ret or u'<abandoned private channel>'

	def __unicode__(self):
		return u'NChannel {0}'.format(get_name())

	def __str__(self):
		return 'NChannel {0}'.format(get_name())

	def __repr__(self):
		return 'NChannel({0}, {1})'.format(self.name, repr(self.topic))

	def __lt__(self, other):
		return self.name < other.name



class UChannel(object):
	def __init__(self, user, nchan, alias=None):
		self.user = user        # the user which this object belongs to
		self.nchan = nchan      # the NChannel object
		self.alias = alias      # local channel name (private)
		self.last_read = 0      # most recent sno viewed in this channel
		self.last_ping = 0      # most recent sno that was a hilight
		self.hilights = False
		self.activity = False
		self.display_notification = False
		self.lock_to_bottom = True
		self.vis = []           # visible messages

	def update_activity_flags(self, set_last_read=False, last_nchan_msg=0):
		if set_last_read:
			if self.vis:
				self.last_read = max(self.last_read, self.vis[-1].msg.sno)
			else:
				self.last_read = 0
		
		if not last_nchan_msg and self.nchan.msgs:
			last_nchan_msg = self.nchan.msgs[-1].sno

		hilights = self.last_read < self.last_ping
		activity = self.last_read < last_nchan_msg

		self.display_notification = hilights or activity
		
		if self.display_notification \
		and self == self.user.active_chan \
		and self.lock_to_bottom:
			# don't display notifications in the status bar
			# when chan is active and bottom messages are visible
			self.display_notification = False
	
		if hilights and hilights != self.hilights:
			self.user.client.notify_new_hilight(self)

		if self.hilights != hilights \
		or self.activity != activity :
			self.hilights = hilights
			self.activity = activity
			return True
		
		return False

	def jump_to_msg(self, msg_n):
		if msg_n >= len(self.nchan.msgs):
			msg_n = len(self.nchan.msgs)-1
		
		self.vis = [VisMessage().c_new(
			self.nchan.msgs[msg_n], [u'x'], msg_n, 0, 1, self)]
		
		self.lock_to_bottom = False
		self.user.client.need_full_redraw = True

	def jump_to_time(self, dt):
		ts = calendar.timegm(dt.timetuple())
		for msg in self.nchan.msgs:
			if msg.ts >= ts:
				i = self.nchan.msgs.index(msg)
				print('jump to {0} of {1}'.format(i, len(self.nchan.msgs)))
				return self.jump_to_msg(i)
		
		return self.jump_to_msg(len(self.nchan.msgs)-1)



class VisMessage(object):
	def __init__(self):
		pass

	def c_new(self, msg, txt, im, car, cdr, ch):
		self.msg = msg          # the message object
		self.txt = txt          # the formatted text
		self.im  = im           # offset into the channel's message list
		self.car = car          # first visible line
		self.cdr = cdr          # last visible line PLUS ONE
		self.vt100 = ch.user.client.vt100

		if not msg or not msg.user: whoops('msg bad')
		if not ch  or not ch.user:  whoops('user bad')
		
		self.unformatted = txt[0]
		self.hilight = bool(ch.user.nick_re.search(msg.txt))
		self.unread = \
			msg.user != ch.user.nick \
			and msg.sno > ch.last_read

		#print('add msg for {0} which is unread {1}, hilight {2}'.format(
		#	ch.user.nick, self.unread, self.hilight))

		self.apply_markup()
		return self

	def c_segm(self, other, src_car, src_cdr, new_car, new_cdr, ch):
		self.msg = other.msg
		self.txt = other.txt[src_car:src_cdr]
		self.im  = other.im
		self.car = new_car
		self.cdr = new_cdr
		self.vt100 = ch.user.client.vt100

		self.hilight = other.hilight
		self.unread = other.unread
		if src_car == 0:
			self.unformatted = other.unformatted
		else:
			self.unformatted = self.txt[0]

		return self

	def plaintext():
		return [self.unformatted] + self.txt[1:]

	def apply_markup(self):
		if self.vt100:
			postfix = u'\033[0m '
			if self.hilight and self.unread:
				prefix = u'\033[1;35;7m'
			elif self.hilight:
				prefix = u'\033[1;35m'
			elif self.unread:
				prefix = u'\033[7m'
			else:
				prefix = u''
				postfix = None
		else:
			prefix = u''
			if self.hilight:
				postfix = u'='
			else:
				postfix = None

		if postfix and not self.unformatted.startswith(u' '):
			#print('applying prefix {0}'.format(b2hex(prefix.encode('utf-8'))))
			ofs = self.unformatted.find(u' ')
			self.txt[0] = u'{0}{1}{2}{3}'.format(
				prefix, self.unformatted[:ofs], \
				postfix, self.unformatted[ofs+1:])
		else:
			self.txt[0] = self.unformatted



class Message(object):
	def __init__(self, to, ts, user, txt):
		self.ts   = ts          # int timestamp
		self.user = user        # str username
		self.txt  = txt         # str text
		
		# set serial number based on last message in target
		if to and to.msgs:
			self.sno = to.msgs[-1].sno + 1
		else:
			self.sno = 0

	def __unicode__(self):
		hhmmss = datetime.datetime.utcfromtimestamp(self.ts).strftime('%Y-%m%d-%H%M%S')
		return u'Message {0:x} time({1},{2}) from({3}) body({4})'.format(
			id(self), self.ts, hhmmss, self.user, self.txt)

	def __str__(self):
		hhmmss = datetime.datetime.utcfromtimestamp(self.ts).strftime('%Y-%m%d-%H%M%S')
		return 'Message {0:x} time({1},{2}) from({3}) body({4})'.format(
			id(self), self.ts, hhmmss, self.user, self.txt)

	def __repr__(self):
		return 'Message({0}, \'{1}\', {2})'.format(self.ts, self.user, repr(self.txt))

